# -*- coding: utf-8 -*-
"""`CalcJob` implementation for DFTK."""
import io
import os
import json
import typing as ty

from aiida import orm
from aiida.common import datastructures, exceptions
from aiida.engine import CalcJob, ExitCode
from aiida_pseudo.data.pseudo import UpfData
from pymatgen.core import units


_AIIDA_DFTK_VERSION_SPEC = "0.2.0"


class DftkCalculation(CalcJob):
    """`CalcJob` implementation for DFTK."""

    INPUT_FILENAME = 'run_dftk.json'
    LOGFILE = 'run_dftk.log'
    SCFRES_SUMMARY_NAME = 'self_consistent_field.json'
    # TODO: don't limit postscf
    _SUPPORTED_POSTSCF = ['compute_forces_cart', 'compute_stresses_cart', 'compute_bands']
    _PSEUDO_SUBFOLDER = './pseudo/'
    _MIN_OUTPUT_BUFFER_TIME = 60

    @staticmethod
    def _merge_dicts(dict1, dict2):
        """Recursively merge dict2 into dict1."""
        for key, value in dict2.items():
            if key in dict1 and isinstance(dict1[key], dict) and isinstance(value, dict):
                DftkCalculation._merge_dicts(dict1[key], value)
            else:
                dict1[key] = value

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        super().define(spec)
        # Inputs
        spec.input('structure', valid_type=orm.StructureData, help='structure')
        spec.input_namespace('pseudos', valid_type=UpfData, help='The pseudopotentials.', dynamic=True)
        spec.input('pseudo_rcut', valid_type=orm.Float, required=False, help='rcut to use for the pseudopotentials, in atomic units')
        spec.input('kpoints', valid_type=orm.KpointsData, help='kpoint mesh or kpoint path')
        spec.input('parameters', valid_type=orm.Dict, help='input parameters')
        spec.input('parent_folder', valid_type=orm.RemoteData, required=False, help='A remote folder used for restarts.')

        options = spec.inputs['metadata']['options']

        options['parser_name'].default = 'dftk'
        options['input_filename'].default = cls.INPUT_FILENAME
        options['max_wallclock_seconds'].default = 1800

        # TODO: Why is this here?
        options['resources'].default = {'num_machines': 1, 'num_mpiprocs_per_machine': 1}
        options['withmpi'].default = True

        # Exit codes
        # TODO: Codes 1xx are already used in the super class!
        spec.exit_code(101, 'ERROR_MISSING_SCFRES_FILE', message='The output file containing SCF results is missing.')
        spec.exit_code(102, 'ERROR_MISSING_FORCES_FILE', message='The output file containing forces is missing.')
        spec.exit_code(103, 'ERROR_MISSING_STRESSES_FILE', message='The output file containing stresses is missing.')
        spec.exit_code(104, 'ERROR_MISSING_BANDS_FILE',message='The output file containing bands is missing.')
        spec.exit_code(500, 'ERROR_SCF_CONVERGENCE_NOT_REACHED', message='The SCF minimization cycle did not converge, and the POSTSCF functions were not executed.')
        spec.exit_code(501, 'ERROR_SCF_OUT_OF_WALLTIME',message='The SCF was interuptted due to out of walltime. Non-recovarable error.')
        spec.exit_code(502, 'ERROR_POSTSCF_OUT_OF_WALLTIME',message='The POSTSCF was interuptted due to out of walltime.')
        spec.exit_code(503, 'ERROR_BANDS_CONVERGENCE_NOT_REACHED', message='The BANDS minimization cycle did not converge.')
        # Significant errors but calculation can be used to restart
        spec.exit_code(400, 'ERROR_PACKAGE_IMPORT_FAILED', message="Failed to import AiiDA DFTK or write first log message. Typically indicates an environment issue.")

        # Outputs
        spec.output('output_parameters', valid_type=orm.Dict, help='output parameters')
        # TODO: doesn't seem to be used?
        spec.output('output_structure', valid_type=orm.Dict, required=False, help='output structure')
        # TODO: doesn't seem to be used?
        spec.output(
            'output_kpoints', valid_type=orm.KpointsData, required=False, help='kpoints array, if generated by DFTK'
        )
        spec.output('output_forces', valid_type=orm.ArrayData, required=False, help='forces array')
        spec.output('output_stresses', valid_type=orm.ArrayData, required=False, help='stresses array')
        spec.output('output_bands', valid_type=orm.BandsData, required=False, help='bandstructure')

        # TODO: bands and DOS implementation required on DFTK side
        # spec.output('output_bands', valid_type=orm.BandsData, required=False,
        #     help='eigenvalues array')
        # spec.output('output_dos')

        spec.default_output_node = 'output_parameters'

    def _validate_options(self):
        """Validate the options input.
        
        Check that the wihmpi option is set to True if the number of mpiprocs is greater than 1.
        Check max_wallclock_seconds is greater than the min_output_buffer_time.
        """
        options = self.inputs.metadata.options
        if options.withmpi is False and options.resources.get('num_mpiprocs_per_machine', 1) > 1:
            # TODO: does aiida not already check this?
            raise exceptions.InputValidationError('MPI is required when num_mpiprocs_per_machine > 1.')
        if options.max_wallclock_seconds < self._MIN_OUTPUT_BUFFER_TIME:
            raise exceptions.InputValidationError(
                f'max_wallclock_seconds must be greater than {self._MIN_OUTPUT_BUFFER_TIME}.'
            )

    def _validate_inputs(self):
        """Validate input parameters."""
        parameters = self.inputs.parameters.get_dict()
        if 'postscf' in parameters:
            for postscf in parameters['postscf']:
                if postscf['$function'] not in self._SUPPORTED_POSTSCF:
                    raise exceptions.InputValidationError(f"Unsupported postscf function: {postscf['$function']}")

        # We want the option to be set for `verdi calcjob inputcat` to work,
        # but we don't allow overriding it because it would affect the name of the log file.
        if self.metadata.options.input_filename != self.INPUT_FILENAME:
            raise exceptions.InputValidationError(
                f"Input filename must be {self.INPUT_FILENAME}. Was: {self.metadata.options.input_filename}"
            )

    def _validate_pseudos(self):
        """Validate the pseudopotentials.

        Check that there is a one-to-one map of kinds in the structure to pseudopotentials.
        """
        kinds = set(kind.name for kind in self.inputs.structure.kinds)
        pseudos = set(self.inputs.pseudos.keys())
        if kinds != pseudos:
            raise exceptions.InputValidationError(
                'Mismatch between the defined pseudos and the list of kinds of the structure.\n'
                f'Pseudos: {pseudos};\nKinds:{kinds}'
            )

    def _validate_kpoints(self):
        """Validate the k-points intput.

        Check that the input k-points provide a k-points mesh.
        """
        try:
            self.inputs.kpoints.get_kpoints_mesh()
        except AttributeError as exc:
            raise exceptions.InputValidationError('The kpoints input does not have a valid mesh set.') from exc

    def _generate_inputdata(
        self, parameters: orm.Dict, structure: orm.StructureData, pseudos: dict, kpoints: orm.KpointsData
    ) -> ty.Tuple[dict, list]:
        """Generate the input dict (json) for DFTK.

        :param parameters: a dict defines the calculation parameters for DFTK
        :param structre: a StructureDate define the crystal
        :param pseudos: a dict contains the pseudos
        :param kpoints: a KpointData
        :return: dict for the DFTK json input
        :return: list of a pseudos needed to be copied
        """

        local_copy_pseudo_list = []

        data = {
            'periodic_system': {},
            'pseudopotentials': {},
            'model_kwargs': {},
            'basis_kwargs': {},
            'scf': {},
            'postscf': [],
        }
        data['periodic_system']['bounding_box'] = [[x * units.ang_to_bohr for x in vec] for vec in structure.cell]
        data['periodic_system']['atoms'] = []
        for site in structure.sites:
            data['periodic_system']['atoms'].append({
                'symbol': site.kind_name,
                'position': [X * units.ang_to_bohr for X in list(site.position)],
            })
        for symbol, pseudo in pseudos.items():
            data['pseudopotentials'][symbol] = f'{self._PSEUDO_SUBFOLDER}{pseudo.filename}'
            local_copy_pseudo_list.append((pseudo.uuid, pseudo.filename, f'{self._PSEUDO_SUBFOLDER}{pseudo.filename}'))
        if 'pseudo_rcut' in self.inputs:
            data['pseudopotentials']['$kwargs'] = {'rcut': self.inputs.pseudo_rcut.value}
        data['basis_kwargs']['kgrid'], data['basis_kwargs']['kshift'] = kpoints.get_kpoints_mesh()

        # set the maxtime for the SCF cycle, with a margin of _MIN_OUTPUT_BUFFER_TIME and 10%, whichever leads to a larger margin
        maxtime = min(
            self.inputs.metadata.options.max_wallclock_seconds - self._MIN_OUTPUT_BUFFER_TIME,
            0.9 * self.inputs.metadata.options.max_wallclock_seconds,
        )
        data['scf']['maxtime'] = maxtime
        
        DftkCalculation._merge_dicts(data, parameters.get_dict())

        return data, local_copy_pseudo_list

    def _generate_retrieve_list(self, parameters: orm.Dict) -> list:
        """Generate the list of files to retrieve based on the type of calculation requested in the input parameters.

        :param parameters: input parameters
        :returns: list of files to retreive
        """
        parameters = parameters.get_dict()
        # Retrieve the postscf files, all function.hdf5 except compute_bands.json
        retrieve_list = [
            f"{item['$function']}.json" if item['$function'] == 'compute_bands' else f"{item['$function']}.hdf5"
            for item in parameters['postscf']
        ]
        retrieve_list.append(self.LOGFILE)
        retrieve_list.append('timings.json')
        retrieve_list.append(f'{self.SCFRES_SUMMARY_NAME}')
        return retrieve_list

    def prepare_for_submission(self, folder):
        """Create the input file(s) from the input nodes.

        :param folder: an `aiida.common.folders.Folder` where the plugin should temporarily place all files needed by
            the calculation.
        :return: `aiida.common.datastructures.CalcInfo` instance
        """

        self._validate_options()
        self._validate_inputs()
        self._validate_pseudos()
        self._validate_kpoints()

        # Create lists which specify files to copy and symlink
        remote_copy_list = []
        remote_symlink_list = []

        # Generate the input file content
        input_filecontent, local_copy_list = self._generate_inputdata(self.inputs.parameters, self.inputs.structure, self.inputs.pseudos, self.inputs.kpoints)

        # write input file
        input_filename = folder.get_abs_path(self.metadata.options.input_filename)
        with io.open(input_filename, 'w', encoding='utf-8') as stream:
            json.dump(input_filecontent, stream, indent=4)

        # List the files (scfres.jld2) to copy or symlink in the case of a restart
        if 'parent_folder' in self.inputs:
            # Symlink if on the same computer, otherwise copy
            same_computer = self.inputs.code.computer.uuid == self.inputs.parent_folder.computer.uuid
            checkpointfile_info = (
                self.inputs.parent_folder.computer.uuid,
                os.path.join(self.inputs.parent_folder.get_remote_path(), self.inputs.parameters['scf']['checkpointfile']),
                self.inputs.parameters['scf']['checkpointfile']
            )
            if same_computer:
                remote_symlink_list.append(checkpointfile_info)
            else:
                remote_copy_list.append(checkpointfile_info)

        # prepare command line parameters
        cmdline_params = [
            # Precompilation under MPI generally deadlocks. Make sure everything is already precompiled.
            '--compiled-modules=strict',
            '-e', 'using AiidaDFTK; AiidaDFTK.run(inputfile="{}", allowed_versions="{}")'.format(
                self.metadata.options.input_filename,
                _AIIDA_DFTK_VERSION_SPEC,
            ),
        ]

        # prepare retrieve list
        retrieve_list = self._generate_retrieve_list(self.inputs.parameters)

        # Set up the `CodeInfo` to pass to `CalcInfo`
        codeinfo = datastructures.CodeInfo()
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.cmdline_params = cmdline_params

        # Set up the `CalcInfo` so AiiDA knows what to do with everything
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.retrieve_list = retrieve_list
        calcinfo.remote_symlink_list = remote_symlink_list
        calcinfo.remote_copy_list = remote_copy_list
        calcinfo.local_copy_list = local_copy_list

        return calcinfo

class PrecompileCalculation(CalcJob):
    """Calcjob implementation to precompile AiidaDFTK."""

    _SUCCESS_PRINT = "Import successful"

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        super().define(spec)

        options = spec.inputs['metadata']['options']
        options['resources'].default = {'num_machines': 1, 'num_mpiprocs_per_machine': 1}

    def prepare_for_submission(self, folder):
        # Set up the `CodeInfo` to pass to `CalcInfo`
        codeinfo = datastructures.CodeInfo()
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.cmdline_params = [
            '-e', f'using Pkg; Pkg.precompile(; strict=true); using AiidaDFTK; println("{self._SUCCESS_PRINT}")'
        ]
        codeinfo.withmpi = False

        # Set up the `CalcInfo` so AiiDA knows what to do with everything
        calcinfo = datastructures.CalcInfo()
        calcinfo.codes_info = [codeinfo]

        return calcinfo

    # Easier to override the parse method than to write a parser.
    def parse(self, *args, **kwargs):
        exit_code = super().parse(*args, **kwargs)
        if exit_code.status != 0:
            return exit_code

        retrieved = self.node.outputs.retrieved
        filename_stdout = self.node.get_option('scheduler_stdout')

        if filename_stdout is None:
            self.report('could not determine `stdout` filename because `scheduler_stdout` option was not set.')
        else:
            try:
                scheduler_stdout = retrieved.base.repository.get_object_content(filename_stdout, mode='r')
                if self._SUCCESS_PRINT in scheduler_stdout:
                    return ExitCode(0)
            except FileNotFoundError:
                self.report(f'could not parse scheduler output: the `{filename_stdout}` file is missing')

        return self.exit_codes.ERROR_UNSPECIFIED



# -*- coding: utf-8 -*-
"""aiida-abinit utility functions."""

#from .dictionary import *
from .kpoints import *
from .pseudos import *
from .seekpath import *

#from .resources import *

__all__ = kpoints.__all__ + pseudos.__all__ + seekpath.__all__ # pylint: disable=undefined-variable



# -*- coding: utf-8 -*-
"""Pseudopotential utility functions."""
import typing as ty

from aiida import orm
from aiida_pseudo.data.pseudo import UpfData

__all__ = ('validate_and_prepare_pseudos_inputs',)


def validate_and_prepare_pseudos_inputs(
    structure: orm.StructureData,
    pseudos: ty.Optional[ty.Dict[str, UpfData]] = None  # pylint: disable=unsubscriptable-object
) -> ty.Dict[str, UpfData]:  # pylint: disable=invalid-name
    """Validate the given pseudos mapping with respect to the given structure.

    The pseudos dictionary should now be a dictionary of UpfData nodes with the kind as linkname
    As such, if there are multiple kinds with the same element, there will be duplicate UpfData nodes
    but multiple links for the same input node are not allowed. Moreover, to couple the UpfData nodes
    to the Calculation instance, we have to go through the use_pseudo method, which takes the kind
    name as an additional parameter. When creating a Calculation through a Process instance, one
    cannot call the use methods directly but rather should pass them as keyword arguments. However,
    we can pass the additional parameters by using them as the keys of a dictionary

    :param structure: StructureData node
    :param pseudos: a dictionary where keys are the kind names and value are Psp8 nodes
    :raises: ValueError if no Psp8 is found for every element in the structure
    :returns: a dictionary of Psp8 nodes where the key is the kind name
    """

    if isinstance(pseudos, (str, orm.Str)):
        raise TypeError('you passed "pseudos" as a string - maybe you wanted to pass it as "pseudo_family" instead?')

    for kind in structure.get_kind_names():
        if kind not in pseudos:
            raise ValueError(f'no pseudo available for element {kind}')
        elif not isinstance(pseudos[kind], UpfData):
            raise ValueError(f'pseudo for element {kind} is not of type UpfData')

    return pseudos



# -*- coding: utf-8 -*-
"""k-point utility functions."""
from aiida import orm
from aiida.engine import calcfunction
import numpy as np

__all__ = ('create_kpoints_from_distance',)


@calcfunction
def create_kpoints_from_distance(structure: orm.StructureData, distance: orm.Float) -> orm.KpointsData:
    """Generate a uniformly spaced kpoint mesh for a given structure.

    The spacing between kpoints in reciprocal space is guaranteed to be at least the defined distance.

    :param structure: the StructureData to which the mesh should apply
    :param distance: a Float with the desired distance between kpoints in reciprocal space
    :returns: a KpointsData with the generated mesh
    """
    epsilon = 1E-5

    kpoints = orm.KpointsData()
    kpoints.set_cell_from_structure(structure)
    kpoints.set_kpoints_mesh_from_density(distance.value)

    lengths_vector = [np.linalg.norm(vector) for vector in structure.cell]
    lengths_kpoint = kpoints.get_kpoints_mesh()[0]

    is_symmetric_cell = all(abs(length - lengths_vector[0]) < epsilon for length in lengths_vector)
    is_symmetric_mesh = all(length == lengths_kpoint[0] for length in lengths_kpoint)

    # If the vectors of the cell all have the same length, the kpoint mesh should be isotropic as well
    if is_symmetric_cell and not is_symmetric_mesh:
        nkpoints = max(lengths_kpoint)
        kpoints.set_kpoints_mesh([nkpoints, nkpoints, nkpoints])

    return kpoints



"""DFTK bands WorkChain implementation."""

from aiida.engine import WorkChain, ToContext, if_
from aiida.orm import BandsData


from aiida import orm
from aiida.common import AttributeDict, exceptions
from aiida.engine import BaseRestartWorkChain, ProcessHandlerReport, process_handler, while_
from aiida.plugins import CalculationFactory

from aiida_dftk.utils import create_kpoints_from_distance, validate_and_prepare_pseudos_inputs, seekpath_structure_analysis

from aiida_dftk.workflows.base import DftkBaseWorkChain

DftkCalculation = CalculationFactory('dftk')



class DftkBandsWorkChain(WorkChain):
    """ DFTK Bands Workchain to perform a DFT calculation. Validates parameters and restart."""

    _process_class = DftkCalculation

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)

        spec.input('bands_kpoints', valid_type=orm.KpointsData, required=False,
            help='Explicit kpoints to use for the BANDS calculation. Specify either this or `bands_kpoints_distance`.')
        spec.input('bands_kpoints_distance', valid_type=orm.Float, required=False,
            help='Minimum kpoints distance for the BANDS calculation. Specify either this or `bands_kpoints`.')
        spec.expose_inputs(DftkBaseWorkChain, namespace='dftk_base')

        spec.outline(
            cls.setup,
            # cls.validate_parameters,
            # cls.validate_kpoints,
            # cls.validate_pseudos,
            # cls.validate_resources,
            if_(cls.should_run_seekpath)(
                cls.run_seekpath,
            ),
            # cls.prepare_process,
            cls.run_process,
            # cls.inspect_process,
            cls.results,
        )

        # spec.expose_outputs(DftkCalculation)
        spec.output('band_parameters', valid_type=orm.Dict,
            required=True,
            help='The parameters used in the DftkBaseWorkChain.')
        spec.output('band_structure', valid_type=BandsData,
            required=True,
            help='The band structure data of the final calculation.')
        spec.output('seekpath_parameters', valid_type=orm.Dict,
            required=False,
            help='The parameters used in the SeeKpath call to normalize the input or relaxed structure.')


        spec.exit_code(301, 'ERROR_INVALID_INPUT_KPATH',
            message='Neither the `bands_kpoints` nor the `bands_kpoints_distance` input was specified, or both were specified.')

    def setup(self):
        """ create the inputs dictionary in `self.ctx.inputs`. """

        self.ctx.inputs = AttributeDict(self.exposed_inputs(DftkBaseWorkChain, 'dftk_base'))

    def validate_kpath(self):
        """Validate the inputs related to k-points.
        Either an explicit `KpointsData` with given path, or a desired kpath distance should be specified. 
        """

        # if both specified or both not specified
        if ('bands_kpoints' in self.inputs) == ('bands_kpoints_distance' in self.inputs):  # Both are present or both are absent
            return self.exit_codes['ERROR_INVALID_INPUT_KPOINTS']

        
    def should_run_seekpath(self):
        """Seekpath should only be run if the `bands_kpoints` input is not specified."""
        return 'bands_kpoints' not in self.inputs

    def run_seekpath(self):
        """Run the structure through SeeKpath to get the normalized structure and path along high-symmetry k-points .
        This is only called if the `bands_kpoints` input was not specified.
        """
        inputs = {
            'reference_distance': self.inputs.get('bands_kpoints_distance', None),
            'metadata': {
                'call_link_label': 'seekpath'
            }
        }
        result = seekpath_structure_analysis(self.ctx.inputs.dftk.structure, **inputs)
        self.ctx.bands_kpoints = result['explicit_kpoints']

        if 'parameters' in result:
            self.out('seekpath_parameters', result['parameters'])
        else:
            self.report('No seekpath parameters found')

    def run_process(self):
        """
        Run the DftkBaseWorkChain to perform the DFT calculation.
        Ensures 'compute_bands' is appropriately added or updated in 'postscf'.
        """
        inputs = AttributeDict(self.exposed_inputs(DftkBaseWorkChain, namespace='dftk_base'))
        original_parameters = inputs.dftk.parameters.get_dict()
        postscf_operations = original_parameters.get('postscf', [])

        # Update or add 'compute_bands' in postscf operations
        for operation in postscf_operations:
            if operation.get('$function') == 'compute_bands':
                operation['$kwargs'] = {'kpath': self.ctx.bands_kpoints.get_kpoints().tolist()}
                break
        else:
            postscf_operations.append({
                '$function': 'compute_bands',
                '$kwargs': {'kpath': self.ctx.bands_kpoints.get_kpoints().tolist()}
            })

        # Update and submit with new parameters
        new_parameters = dict(original_parameters, postscf=postscf_operations)
        inputs.dftk.parameters = orm.Dict(dict=new_parameters)
        running = self.submit(DftkBaseWorkChain, **inputs)
        self.report(f'Launching DftkBaseWorkChain<{running.pk}>')
        return ToContext(workchain_bands=running)

    def results(self):
        """Attach the desired output nodes directly as outputs of the workchain."""
        self.report('workchain succesfully completed')
        if 'output_parameters' in self.ctx.workchain_bands.outputs:
            self.out('band_parameters', self.ctx.workchain_bands.outputs.output_parameters)
        else:
            self.report('No output parameters found in workchain_bands.outputs')
        if 'output_bands' in self.ctx.workchain_bands.outputs:
            self.out('band_structure', self.ctx.workchain_bands.outputs.output_bands)
        else:
            self.report('No output bands found in workchain_bands.outputs')


# -*- coding: utf-8 -*-
"""Base DFTK WorkChain implementation."""
from aiida import orm
from aiida.common import AttributeDict, exceptions
from aiida.engine import BaseRestartWorkChain, ProcessHandlerReport, process_handler, while_, if_, ToContext
from aiida.plugins import CalculationFactory

from aiida_dftk.utils import create_kpoints_from_distance, validate_and_prepare_pseudos_inputs

DftkCalculation = CalculationFactory('dftk')
PrecompileCalculation = CalculationFactory('dftk.precompile')


class DftkBaseWorkChain(BaseRestartWorkChain):
    """Base DFTK Workchain to perform a DFT calculation. Validates parameters and restart."""

    _process_class = DftkCalculation

    _attempted_precompilation_extra = "attempted_precompilation"

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)

        spec.input('kpoints',
                   valid_type=orm.KpointsData,
                   required=False,
                   help='An explicit k-points mesh or list. Either this or `kpoints_distance` '
                        'must be provided.')
        spec.input('kpoints_distance',
                   valid_type=orm.Float,
                   required=False,
                   help='The minimum desired distance in 1/Å between k-points in reciprocal space. The explicit '
                        'k-point mesh will be generated automatically by a calculation function based on the input '
                        'structure.')
        spec.expose_inputs(DftkCalculation,
                           namespace='dftk',
                           exclude=('kpoints',))

        spec.outline(
            cls.setup,
            cls.validate_kpoints,
            cls.validate_pseudos,
            cls.validate_resources,
            while_(cls.should_run_process)(
                cls.prepare_process,
                cls.run_process,
                if_(cls.should_attempt_precompilation)(
                    cls.run_precompilation,
                    cls.inspect_precompilation,
                ).else_(
                    cls.inspect_process,
                )
            ),
            cls.results,
        )

        spec.expose_outputs(DftkCalculation)

        spec.exit_code(201, 'ERROR_INVALID_INPUT_PSEUDO_POTENTIALS',
            message='`pseudos` could not be used to get the necessary pseudos.')
        spec.exit_code(202, 'ERROR_INVALID_INPUT_KPOINTS',
            message='Neither the `kpoints` nor the `kpoints_distance` input was specified.')
        spec.exit_code(203, 'ERROR_INVALID_INPUT_RESOURCES',
            message='Neither the `options` nor `automatic_parallelization` input was specified.')
        spec.exit_code(204, 'ERROR_INVALID_INPUT_RESOURCES_UNDERSPECIFIED',
            message='The `metadata.options` did not specify both `resources.num_machines` and `max_wallclock_seconds`.')
        spec.exit_code(300, 'ERROR_PRECOMPILATION_FAILURE',
            message='Failed to precompile AiidaDFTK. Typically indicates an environment issue.')

    def setup(self):
        """Call the `setup` of the `BaseRestartWorkChain` and then create the inputs dictionary in `self.ctx.inputs`.

        This `self.ctx.inputs` dictionary will be used by the `BaseRestartWorkChain` to submit the calculations in the
        internal loop.
        """
        super().setup()
        self.ctx.restart_calc = None
        self.ctx.inputs = AttributeDict(self.exposed_inputs(DftkCalculation, 'dftk'))

    # TODO: We probably want to handle the kpoint distance on the Julia side instead.
    def validate_kpoints(self):
        """Validate the inputs related to k-points.

        Either an explicit `KpointsData` with given mesh/path, or a desired k-points distance should be specified. In
        the case of the latter, the `KpointsData` will be constructed for the input `StructureData` using the
        `create_kpoints_from_distance` calculation function.
        """

        # if both specified or both not specified
        if ('kpoints' in self.inputs) == ('kpoints_distance' in self.inputs):  # Both are present or both are absent
            return self.exit_codes['ERROR_INVALID_INPUT_KPOINTS']


        try:
            kpoints = self.inputs.kpoints
        except AttributeError:
            inputs = {
                'structure': self.inputs.dftk.structure,
                'distance': self.inputs.kpoints_distance,
                'metadata': {'call_link_label': 'create_kpoints_from_distance'}
            }
            kpoints = create_kpoints_from_distance(**inputs)  # pylint: disable=unexpected-keyword-arg

        self.ctx.inputs.kpoints = kpoints

    def validate_pseudos(self):
        """Validate the inputs related to pseudopotentials.

        The pseudo potentials should be defined explicitly in the `pseudos` namespace
        """
        structure = self.ctx.inputs.structure
        pseudos = self.inputs.dftk.get('pseudos', None)

        try:
            self.ctx.inputs.pseudos = validate_and_prepare_pseudos_inputs(structure, pseudos)
        except ValueError as exception:
            self.report(f'{exception}')
            return self.exit_codes.ERROR_INVALID_INPUT_PSEUDO_POTENTIALS  # pylint: disable=no-member


    # TODO: This is weird. Shouldn't aiida already handle this internally?
    def validate_resources(self):
        """Validate the inputs related to the resources.

        `metadata.options` should at least contain the options `resources` and `max_wallclock_seconds`,
        where `resources` should define the `num_machines`.
        """
        num_machines = self.ctx.inputs.metadata.options.get('resources', {}).get('num_machines', None)
        max_wallclock_seconds = self.ctx.inputs.metadata.options.get('max_wallclock_seconds', None)

        if num_machines is None or max_wallclock_seconds is None:
            return self.exit_codes.ERROR_INVALID_INPUT_RESOURCES_UNDERSPECIFIED  # pylint: disable=no-member


    def prepare_process(self):
        """Prepare the inputs for the next calculation.

        If a `restart_calc` has been set in the context, its `remote_folder` will be used as the `parent_folder` input
        for the next calculation and the `restart_mode` is set to `restart`. Otherwise, no `parent_folder` is used and
        `restart_mode` is set to `from_scratch`.
        """

        # AiidaDFTK will automatically check the existance of a checkpoint(scfres.jld2) and restart from it
        if self.ctx.restart_calc:
            self.ctx.inputs.parent_folder = self.ctx.restart_calc.outputs.remote_folder

    def report_error_handled(self, calculation, action):
        """Report an action taken for a calculation that has failed.

        This should be called in a registered error handler if its condition is met and an action was taken.

        :param calculation: the failed calculation node
        :param action: a string message with the action taken
        """
        self.report(f'{calculation.process_label}<{calculation.pk}> failed with exit status {calculation.exit_status}: {calculation.exit_message}')
        self.report(f'action taken: {action}')

    def should_attempt_precompilation(self):
        if self.node.base.extras.get(self._attempted_precompilation_extra, False):
            return False
        node = self.ctx.children[self.ctx.iteration - 1]
        return node.exit_code == DftkCalculation.exit_codes.ERROR_PACKAGE_IMPORT_FAILED

    def run_precompilation(self):
        self.node.base.extras.set(self._attempted_precompilation_extra, True)
        calculation = self.ctx.children[self.ctx.iteration - 1]
        self.report_error_handled(calculation, 'attempting to precompile')

        node = self.submit(PrecompileCalculation, inputs={
            "code": calculation.inputs.code,
        })
        self.report(f'launching {node.process_label}<{node.pk}>')

        return ToContext(precompile_job=node)

    def inspect_precompilation(self):
        calculation = self.ctx.precompile_job

        if calculation.exit_status != 0:
            self.report(f'{calculation.process_label}<{calculation.pk}> failed with exit status {calculation.exit_status}: {calculation.exit_message}')
            self.report(f'the issue can be diagnosed by running `verdi process report {calculation.pk}` and checking the logs.')
            self.report('aborting workchain')
            return self.exit_codes.ERROR_PRECOMPILATION_FAILURE

        self.report(f'{calculation.process_label}<{calculation.pk}> succeeded. Resuming workchain iterations.')
        return None

    @process_handler(priority=500, exit_codes=[DftkCalculation.exit_codes.ERROR_SCF_CONVERGENCE_NOT_REACHED])
    def handle_scf_convergence_not_reached(self, _):
        """Handle unconverged SCF calculations."""
        return None

    # Just as a blueprint, delete after ^ is implemented
    # TODO: What exactly is this doing?
    @process_handler(priority=580, exit_codes=[
        DftkCalculation.exit_codes.ERROR_SCF_CONVERGENCE_NOT_REACHED,
        DftkCalculation.exit_codes.ERROR_POSTSCF_OUT_OF_WALLTIME
        ])
    def handle_recoverable_SCF_unconverged_and_POSTSCF_out_of_walltime_(self, calculation):
        """Handle `ERROR_SCF_CONVERGENCE_NOT_REACHED` and `ERROR_POSTSCF_OUT_OF_WALLTIME` exit code: calculations shut down neatly and we can simply restart."""
        try:
            self.ctx.inputs.structure = calculation.outputs.output_structure
        except exceptions.NotExistent:
            self.ctx.restart_calc = calculation
            self.ctx.inputs.metadata.options.max_wallclock_seconds = 3600
            self.ctx.inputs.parameters['scf']['$kwargs']['maxiter'] = 100
            self.report_error_handled(calculation, 'restart from the last calculation, set max_wallclock_seconds as 3600s, maxiter as 100')
        else:
            self.ctx.restart_calc = None
            self.report_error_handled(calculation, 'out of walltime: structure changed, so restarting from scratch')

        return ProcessHandlerReport(True)



# -*- coding: utf-8 -*-
"""Calcfunction to primitivize a structure and return high symmetry k-point path through its Brillouin zone."""
from aiida.engine import calcfunction
from aiida.orm import Data

__all__ = ('seekpath_structure_analysis',)

@calcfunction
def seekpath_structure_analysis(structure, **kwargs):
    """Primitivize the structure with SeeKpath and generate the high symmetry k-point path through its Brillouin zone.

    This calcfunction will take a structure and pass it through SeeKpath to get the normalized primitive cell and the
    path of high symmetry k-points through its Brillouin zone. Note that the returned primitive cell may differ from the
    original structure in which case the k-points are only congruent with the primitive cell.

    The keyword arguments can be used to specify various Seekpath parameters, such as:

        with_time_reversal: True
        reference_distance: 0.025
        recipe: 'hpkot'
        threshold: 1e-07
        symprec: 1e-05
        angle_tolerance: -1.0

    Note that exact parameters that are available and their defaults will depend on your Seekpath version.
    """
    from aiida.tools import get_explicit_kpoints_path

    # All keyword arugments should be `Data` node instances of base type and so should have the `.value` attribute
    unwrapped_kwargs = {key: node.value for key, node in kwargs.items() if isinstance(node, Data)}

    result = get_explicit_kpoints_path(structure, **unwrapped_kwargs)

    return result




# -*- coding: utf-8 -*-
"""`Parser` implementation for DFTK."""
import json
import pathlib as pl
import numpy as np

from aiida.engine import ExitCode
from aiida.orm import ArrayData, Dict
from aiida.parsers import Parser
from aiida.plugins import DataFactory


from aiida_dftk.calculations import DftkCalculation

import h5py

# DataFactory is used to create the BandsData object
BandsData = DataFactory('core.array.bands')

class DftkParser(Parser):
    """`Parser` implementation for DFTK."""

    # TODO: DEFAULT_ prefix should be removed. I don't think that these names can be changed.
    _DEFAULT_ENERGY_UNIT = 'hartree'
    _DEFAULT_FORCE_FUNCNAME = 'compute_forces_cart'
    _DEFAULT_FORCE_UNIT = 'hartree/bohr'
    _DEFAULT_STRESS_FUNCNAME = 'compute_stresses_cart'
    _DEFAULT_STRESS_UNIT = 'hartree/bohr^3'
    _DEFAULT_BANDS_FUNCNAME = 'compute_bands'
    _DEFAULT_BANDS_UNIT = 'hartree'

    def parse(self, **kwargs):
        """Parse DFTK output files."""
        # if ran_out_of_walltime (terminated illy)
        if self.node.exit_status == DftkCalculation.exit_codes.ERROR_SCHEDULER_OUT_OF_WALLTIME.status:
            # if SCF summary file is not in the list of retrieved files, SCF terminated illy
            if DftkCalculation.SCFRES_SUMMARY_NAME not in self.retrieved.list_object_names():
                return self.exit_codes.ERROR_SCF_OUT_OF_WALLTIME
            # POSTSCF terminated illy
            else:
                return self.exit_codes.ERROR_POSTSCF_OUT_OF_WALLTIME
        
        # Check error file
        try:
            errors_log = self.retrieved.base.repository.get_object_content(DftkCalculation.LOGFILE)
            if "Imports succeeded" not in errors_log:
                return self.exit_codes.ERROR_PACKAGE_IMPORT_FAILED
        except FileNotFoundError:
            return self.exit_codes.ERROR_PACKAGE_IMPORT_FAILED

        if "Finished successfully" not in errors_log:
            return self.exit_codes.ERROR_UNSPECIFIED

        # Check retrieve list to know which files the calculation is expected to have produced.
        try:
            self._parse_optional_result(
                DftkCalculation.SCFRES_SUMMARY_NAME,
                self.exit_codes.ERROR_MISSING_SCFRES_FILE,
                self._parse_output_parameters,
            )

            self._parse_optional_result(
                f'{self._DEFAULT_FORCE_FUNCNAME}.hdf5',
                self.exit_codes.ERROR_MISSING_FORCES_FILE,
                self._parse_output_forces,
            )

            self._parse_optional_result(
                f'{self._DEFAULT_STRESS_FUNCNAME}.hdf5',
                self.exit_codes.ERROR_MISSING_STRESSES_FILE,
                self._parse_output_stresses,
            )

            self._parse_optional_result(
                f'{self._DEFAULT_BANDS_FUNCNAME}.json',
                self.exit_codes.ERROR_MISSING_BANDS_FILE,
                self._parse_output_bands,
            )
        except ParsingFailedException as e:
            return e.exitcode

        return ExitCode(0)

    def _parse_optional_result(self, file_name, missing_file_exitcode, parser):
        # Files passed to the CalcInfo to be retrieved
        retrieve_list = self.node.base.attributes.get('retrieve_list')
        # Files that were actually retrieved
        retrieved_files = self.retrieved.base.repository.list_object_names()

        if file_name in retrieve_list:
            if file_name not in retrieved_files:
                raise ParsingFailedException(missing_file_exitcode)
            with self.retrieved.base.repository.as_path(file_name) as file_path:
                parser(file_path)

    def _parse_output_parameters(self, file_path):
        with open(file_path, 'r', encoding='utf-8') as json_file:
            data = json.load(json_file)

        # Ignore the 'occupation' and 'eigenvalues' keys
        # TODO: add back for bands after implementation in DFTK
        data.pop('occupation', None)
        data.pop('eigenvalues', None)

        # Rename the special keys
        data['norm_delta_rho'] = data.pop('norm_Δρ', None)
        data['fermi_level'] = data.pop('εF', None)

        # Add energy units
        for key in list(data['energies'].keys()):
            unit_key = f'{key}_unit'
            data['energies'][unit_key] = self._DEFAULT_ENERGY_UNIT

        data['fermi_level_unit'] = self._DEFAULT_ENERGY_UNIT

        self.out('output_parameters', Dict(dict=data))
        
        # Check for 'converged'
        if not data.get('converged'):
            return self.exit_codes.ERROR_SCF_CONVERGENCE_NOT_REACHED

        return None

    def _parse_output_forces(self, file_path):
        with h5py.File(file_path, 'r') as h5file:
            force_dict = DftkParser._hdf5_to_dict(h5file)

        # TODO: add a check for the forces array agrees with number of atoms
        force_array = ArrayData()
        force_array.set_array('output_forces', force_dict['results'])
        self.out('output_forces', force_array)
        return None

    def _parse_output_stresses(self, file_path):
        with h5py.File(file_path, 'r') as h5file:
            stress_dict = DftkParser._hdf5_to_dict(h5file)

        stress_array = ArrayData()
        stress_array.set_array('output_stresses', stress_dict['results'])
        self.out('output_stresses', stress_array)
        return None
    
    def _parse_output_bands(self, file_path):
        if not pl.Path(file_path).exists():
            return self.exit_codes.ERROR_MISSING_BANDS_FILE

        with open(file_path, 'r', encoding='utf-8') as json_file:
            bands_dict = json.load(json_file)
        
        if bands_dict['diagonalization']['converged'] is False:
            return self.exit_codes.ERROR_BANDS_CONVERGENCE_NOT_REACHED
        
        bands_data = BandsData()
        kpath = bands_dict['kcoords']
        eigen_array = np.array(bands_dict['eigenvalues'])
        nspin = bands_dict['n_spin_components']
        nkpoints = bands_dict['n_kpoints']
        nbands = bands_dict['n_bands']

        bands = eigen_array.reshape(nspin, nkpoints, nbands)

        bands_data.set_kpoints(kpoints=kpath)
        bands_data.set_bands(bands, units=self._DEFAULT_BANDS_UNIT)
        self.out('output_bands', bands_data)

        return None

    @staticmethod
    def _hdf5_to_dict(hdf5_file):
        """Convert an HDF5 file to a Python dictionary.

        :param hdf5_file: File or group object from h5py (HDF5 file handle or subgroup)
        :return: Dictionary representation of the HDF5 file or group.
        """
        result = {}

        for key, item in hdf5_file.items():
            if isinstance(item, h5py.Dataset):  # item is a dataset
                value = item[()]
                if isinstance(value, bytes):  # Check if the value is bytes and decode if necessary
                    value = value.decode('utf-8')
                result[key] = value
            elif isinstance(item, h5py.Group):  # item is a group
                result[key] = DftkParser._hdf5_to_dict(item)
        return result

class ParsingFailedException(Exception):
    def __init__(self, exitcode: ExitCode):
        super().__init__(exitcode)
        self.exitcode = exitcode
